from typing import Any


class Configuration():
    """A configuration is a selection of elements in a variability model.

    It is represented as a dictionary of elements of Any type -> bool.
    """

    def __init__(self, elements: dict[Any, bool]) -> None:
        self.elements = elements

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Configuration):
            return self.elements == other.elements
        return False

    def __hash__(self) -> int:
        return hash(frozenset(self.elements.items()))

    def __str__(self) -> str:
        return str(self.elements)



from abc import abstractmethod
from typing import Optional

from famapy.core.models import Configuration
from famapy.core.operations import Operation


class Sampling(Operation):
    """Sampling is the selection of a subset (i.e., a sample) of products
    (or configurations) from within a variability model.
    """

    @abstractmethod
    def sample(self, size: int, with_replacement: bool = False, 
               partial_configuration: Optional[Configuration] = None) -> list[Configuration]:
        """Return a sample of configurations.

        Keyword arguments:
        size -- number of configurations of the sample.
        with_replacement -- allow repeated configurations in the sample (default False).
        partial_configuration -- from which the sample is built (default empty configuration).
        """  



from typing import Optional


class Node:
    operations = ['requires', 'excludes', 'and', 'or', 'implies', 'not', 'equivalence']

    def __init__(self, data: str):
        self.left: Optional['Node'] = None  # pylint: disable=unsubscriptable-object
        self.right: Optional['Node'] = None  # pylint: disable=unsubscriptable-object
        self.data = data

    def is_feature(self) -> bool:
        return not self.is_op()

    def is_op(self) -> bool:
        ''' We check if the operation is available by comparing it on mayus '''
        return (
            isinstance(self.data, str) and
            (self.data.upper() in (operation.upper() for operation in Node.operations))
        )

    def __str__(self) -> str:
        if self.left and self.right:
            return f'{self.data}[{self.left}][{self.right}]'

        if self.left and not self.right:
            return f'{self.data}[{self.left}][]'

        if not self.left and self.right:
            return f'{self.data}[][{self.right}]'

        return str(self.data)


class AST:

    def __init__(self, root: Node):
        self.root = root

    @classmethod
    def create_simple_binary_operation(cls, operation: str, left: str, right: str) -> 'AST':
        ast = cls(Node(operation))
        ast.root.left = Node(left)
        ast.root.right = Node(right)
        return ast

    @classmethod
    def create_simple_unary_operation(cls, operation: str, elem: str) -> 'AST':
        ast = cls(Node(operation))
        ast.root.left = Node(elem)
        return ast

    def __str__(self) -> str:
        return str(self.root)



from typing import Any, Optional

import hug

from famapy.core.discover import DiscoverMetamodels
from typing import NewType


dm = DiscoverMetamodels()

PluginsType = NewType('PluginsType', dict[str, list[str]])
OperationDict = NewType('OperationDict', dict[str, list[str]])
OperationResult = NewType('OperationResult', dict[str, Any])


@hug.cli()
@hug.get('/get-plugins/', versions=1)
def get_plugins() -> PluginsType:
    """ Get availables plugins """
    plugins = dm.get_plugins()
    return PluginsType({'plugins': plugins})


@hug.cli()
@hug.get('/get-operations/{plugin}/')
def get_operations_name_by_plugin(plugin: str, versions: int = 1) -> OperationDict:
    """ Get availables operations gave a plugin name """
    operations = dm.get_name_operations_by_plugin(plugin)
    return OperationDict({'operations': operations})


@hug.cli()
def use_operation_from_file(
    operation: str,
    filename: str,
    plugin: Optional[str] = None,
    versions: int = 1
) -> OperationResult:
    """
    Execute an operation gave an operation and one input file. Optionally you
    can give a plugin as last parameter.
    """
    result = dm.use_operation_from_file(operation, filename, plugin)
    return OperationResult({'result': result})
